header {
import stringtemplate
}

header "CMinusParser.__init__" {
    self.main = None
    if kwargs.has_key("main"):
        self.main = kwargs["main"]
    self.currentFunctionName = None
}

header "CMinusParser.__main__" {

import CMinusLexer
import CMinusParser

class Main:

    def __init__(self):
        self.templates = None
        self.templateFileName = "Java.stg";
        if len(sys.argv) > 1:
            self.templateFileName = sys.argv[1]
        self.templates = stringtemplate.StringTemplateGroup( \
                             file(self.templateFileName), \
                             stringtemplate.language.AngleBracketTemplateLexer.Lexer);
        self.lexer = CMinusLexer.Lexer(sys.stdin)
        self.parser = CMinusParser.Parser(self.lexer)
        self.parser.main = self
        try:
            self.code = self.parser.program()
            sys.stdout.write(str(self.code) + '\n')
        except:
            print "parse error in translator"

Main()

}

options {
  language="Python";
}

class CMinusParser extends Parser;

options {
  k=3;
}

{
    def template(self, name):
        return self.main.templates.getInstanceOf(name)

    def text(self, t):
        return stringtemplate.StringTemplate(self.main.templates, t)

}

program returns [code = self.template("program")]
{
    print self.main
}
    :   ( declaration[code] )+
    ;

declaration[code]
{
    f = None
    v = None
}
    :   v = variable {code.setAttribute("globals", v)}
    |   f = function {code.setAttribute("functions", f)}
    ;

variable returns [code = None]
{
    t = None
    d = None
}
    :   t=type d=declarator SEMI
        {
            if not self.currentFunctionName:
                code = self.template("globalVariable")
            else:
                code = self.template("variable")
            code.setAttribute("type", t)
            code.setAttribute("name", d)
        }
    ;

declarator returns [code = None]
    :   id:ID { code = self.text(id.getText()) }
    ;

function returns [code = self.template("function")]
{
    t = None
    f = None
}
    :   t=type id:ID { self.currentFunctionName = id.getText() } LPAREN
        (   f=formalParameter
            { code.setAttribute("args", f) }
            (   COMMA f=formalParameter
                { code.setAttribute("args", f) }
            )*
        )?
        RPAREN
        block[code]
        {
            code.setAttribute("type", t)
            code.setAttribute("name", id.getText())
            self.currentFunctionName = None
        }
    ;

formalParameter returns [code = self.template("parameter")]
{
    t = None
    d = None
}
    :   t=type d=declarator
        {
            code.setAttribute("type", t)
            code.setAttribute("name", d)
        }
    ;

type returns [code = None]
    :   "int"   { code = self.template("type_int") }
    |   "char"  { code = self.template("type_char") }
    |   id:ID
        {
            code = self.template("type_user_object")
            code.setAttribute("name", id.getText())
        }
    ;

block [st]
{
    s = None
    v = None
}
    :   LCURLY
            ( options {warnWhenFollowAmbig=false;} // hush a warning
            : v=variable { st.setAttribute("locals", v) }
            )*
            ( s=stat { st.setAttribute("stats", s) } )*
        RCURLY
    ;

stat returns [code = None]
{
    e = None
}
    : code=forStat
    | e=expr SEMI
      {
          code = self.template("statement")
          code.setAttribute("expr", e)
      }
    | block[code = self.template("statementList")]
    | code=assignStat SEMI
    | SEMI
    ;

forStat returns [code = self.template("forLoop")]
{
    e1 = None
    e2 = None
    e3 = None
    s = None
}
    :   "for" LPAREN e1=assignStat SEMI e2=expr SEMI e3=assignStat RPAREN
        block[code]
        {
            code.setAttribute("e1", e1)
            code.setAttribute("e2", e2)
            code.setAttribute("e3", e3)
        }
    ;

assignStat returns [code = None]
{
     a = None
}
    :   id:ID ASSIGN a=expr
        {
            code = self.template("assign")
            code.setAttribute("lhs", id.getText())
            code.setAttribute("rhs", a)
        }
    ;

expr returns [code = None]
    :   code=condExpr
    ;

condExpr returns [code = None]
{
    a = None
    b = None
    op = None
}
    :   code=aexpr
        (   { op = self.LT(1) }
            (EQUALS | LT) b=aexpr
            {
                a = code
                if op.getType() == EQUALS:
                    code = self.template("equals")
                else:
                    code = self.template("lessThan")
                code.setAttribute("left", a)
                code.setAttribute("right", b)
            }
        )?
    ;

aexpr returns [code = None]
{
    a = None
    b = None
}
    :   code=atom
        (   PLUS b=atom
            {
            a = code
            code = self.template("add")
            code.setAttribute("left", a)
            code.setAttribute("right", b)
            }
        )*
    ;

atom returns [code = None]
    :    id:ID
         {
             code = self.template("refVar")
             code.setAttribute("id", id.getText())
         }
    |    i:INT
         {
             code = self.template("iconst")
             code.setAttribute("value", i.getText());
         }
    |    LPAREN code=expr RPAREN
    ; 

class CMinusLexer extends Lexer;

options {
    charVocabulary = '\1'..'\377';
    k=2;
}

PLUS    :   "+" ;
EQUALS  :   "==" ;
LT      :   '<' ; 
ASSIGN  :   '=' ; 
LPAREN  :   '(' ;
RPAREN  :   ')' ;
LCURLY  :   '{' ;
RCURLY  :   '}' ;
SEMI    :   ';' ; 
STAR    :   '*' ; 
COMMA   :   ',' ;

ID  :   ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*
    ;

INT     :       ('0'..'9')+
        ;

WS  :   (   ' '
        |   '\t'
        |   '\r'
        |   '\n' { $newline }
        )+
        { $skip }
    ;    
